import asyncio
import logging
from typing import List

from asyncpg import LockNotAvailableError
from buildpg.asyncpg import BuildPgConnection

from .. import glove
from ..settings import BaseSettings
from .helpers import DummyPgPool
from .patches import Patch, get_sql_section
from .utils import AsyncPgContext

logger = logging.getLogger('foxglove.db.migrations')

__all__ = ('run_migrations',)


migrations_table_name = 'migrations'
migrations_table_sql = f"""
create table if not exists {migrations_table_name} (
  id serial primary key,
  ts timestamptz not null default current_timestamp,
  ref varchar(255) not null,
  sql_section text not null,
  fake boolean not null,
  unique (ref, sql_section)
);
"""


async def run_migrations(settings: BaseSettings, patches: List[Patch], live: bool, *, fake: bool = False) -> int:
    """
    Migrations in foxglove are handled by patches which are run automatically if
    foxglove spots they haven't been run before
    """
    migration_patches = [p for p in patches if p.auto_run]
    if not migration_patches:
        return 0

    count = 0
    up_to_date = 0
    async with AsyncPgContext(settings.pg_dsn) as conn:
        tr = conn.transaction()
        await tr.start()

        if not await conn.fetchval('select 1 from pg_tables where tablename=$1', migrations_table_name):
            await conn.execute(migrations_table_sql)
            logger.info('%s table created', migrations_table_name)

        try:
            await conn.execute(f'lock table {migrations_table_name} nowait')
        except LockNotAvailableError:
            logger.debug('another transaction has locked %s, skipping migrations here', migrations_table_name)
            await tr.rollback()
            return 0

        default_pg = getattr(glove, 'pg', None)
        glove.pg = DummyPgPool(conn)
        logger.info('checking %d migration patches...', len(migration_patches))
        for patch in migration_patches:
            if patch.auto_sql_section:
                content = get_sql_section(patch.auto_sql_section, settings.sql)
                sql_section = f'{patch.auto_sql_section}::\n{content}'
            else:
                # '-' is required to make the unique constraint work since null would mean rows wouldn't conflict
                sql_section = '-'

            patch_ref = patch.func.__name__
            if isinstance(patch.auto_run, str):
                patch_ref += f':{patch.auto_run}'
            migration_id = await conn.fetchval(
                f"""
                insert into {migrations_table_name} (ref, sql_section, fake)
                values ($1, $2, $3)
                on conflict (ref, sql_section) do nothing
                returning id
                """,
                patch_ref,
                sql_section,
                fake,
            )
            if migration_id is None:
                up_to_date += 1
                continue

            if fake:
                logger.warning('faked migration %s', patch_ref)
            else:
                successful = await run_patch(conn, patch, patch_ref, live)
                if not successful:
                    logger.warning('patch failed, rolling back all %d migration patches in this session', count)
                    await tr.rollback()
                    glove.pg = default_pg
                    return 0

            count += 1

        glove.pg = default_pg
        verb = 'faked' if fake else 'run'
        if live:
            await tr.commit()
            if count == 0:
                logger.info('all %d migrations already up to date ✓', up_to_date)
            else:
                logger.info('%d migration patches %s, %d already up to date ✓', count, verb, up_to_date)
        else:
            await tr.rollback()
            logger.info(
                '%d migration patches %s, %d already up to date, not live rolling back', count, verb, up_to_date
            )

    return count


async def run_patch(conn: BuildPgConnection, patch: Patch, ref: str, live: bool) -> bool:
    kwargs = dict(conn=conn, live=live, args={'__migration__': 'true'}, logger=logger)
    logger.info('{:-^50}'.format(f' running {ref} '))
    try:
        if asyncio.iscoroutinefunction(patch.func):
            result = await patch.func(**kwargs)
        else:
            result = patch.func(**kwargs)
        if result is not None:
            logger.info('result: %s', result)
    except BaseException:
        logger.info('{:-^50}'.format(f' {ref} failed '))
        logger.exception('Error running %s migration patch', patch.func.__name__)
        return False
    else:
        logger.info('{:-^50}'.format(f' {ref} succeeded '))
        return True



import logging
import re
from typing import Optional, Set

from starlette.datastructures import URL
from starlette.requests import Request

from . import exceptions, glove
from .settings import BaseSettings
from .utils import get_ip

logger = logging.getLogger('foxglove.recaptcha')
TESTING_SECRET = '6LeIxAcTAAAAAGG-vFI1TnRWxMZNFuojJ4WifJWe'


async def check_recaptcha(
    request: Request, recaptcha_token: Optional[str], *, allowed_hosts: Set[str] = None, error_headers=None
) -> None:
    client_ip = get_ip(request)

    if not recaptcha_token:
        logger.warning('recaptcha token not provided, path="%s" ip=%s', request.url.path, client_ip)
        raise exceptions.HttpBadRequest('No recaptcha value', headers=error_headers)

    settings: BaseSettings = glove.settings

    post_data = {'secret': settings.recaptcha_secret, 'response': recaptcha_token, 'remoteip': client_ip}
    r = await glove.http.post(settings.recaptcha_url, data=post_data)
    r.raise_for_status()
    data = r.json()

    # use allowed_hosts or settings.origin exists, instead of the request to avoid problems with old browsers
    # that don't include the Origin header
    if allowed_hosts is None:
        if origin := getattr(settings, 'origin', None):
            allowed_hosts = {URL(origin).hostname}
        elif origin := request.headers.get('origin'):
            # using the origin here if available instead of host avoids problems when requests
            # are proxied e.g. with netlify
            allowed_hosts = {re.sub('^https?://', '', origin)}
        else:
            allowed_hosts = {request.url.hostname}

    if data['success']:
        hostname = data['hostname']
        if hostname in allowed_hosts:
            logger.info('recaptcha success')
            return
        elif settings.dev_mode and settings.recaptcha_secret == TESTING_SECRET and hostname == 'testkey.google.com':
            logger.info('recaptcha test key success')
            return

    logger.warning(
        'recaptcha failure, path=%s allowed_hosts=%s ip=%s response=%s',
        request.url.path,
        ','.join(allowed_hosts),
        client_ip,
        r.text,
        extra={'recaptcha_response': data, 'recaptcha_token': recaptcha_token, 'headers': dict(request.headers)},
    )
    raise exceptions.HttpBadRequest('Invalid recaptcha value', headers=error_headers)


class RecaptchaDepends:
    __slots__ = ('request',)

    def __init__(self, request: Request):
        self.request = request

    async def __call__(
        self, recaptcha_token: Optional[str], *, allowed_hosts: Set[str] = None, error_headers=None
    ) -> None:
        return await check_recaptcha(
            self.request, recaptcha_token, allowed_hosts=allowed_hosts, error_headers=error_headers
        )



import asyncio
import hashlib
import secrets
from functools import lru_cache
from time import time
from typing import Optional

import bcrypt
from fastapi import Request
from pydantic import SecretBytes

from . import glove
from .exceptions import HttpTooManyRequests, UnexpectedResponse, manual_response_error

__all__ = 'rate_limit', 'check_password_breached', 'check_password_correct', 'hash_password'


def rate_limit(*, request_limit: Optional[int], interval: int):
    async def check_rate_limit(request: Request) -> int:
        cache_key = f'rate-limit:{request.method}{request.url.path}:{time() // interval:0.0f}'
        with await glove.redis as conn:
            pipe = conn.pipeline()
            pipe.unwatch()
            pipe.incr(cache_key)
            pipe.expire(cache_key, interval)
            _, request_count, _ = await pipe.execute()
            if request_limit is None:
                return request_count or 0
            elif request_count > request_limit:
                raise HttpTooManyRequests(f'rate limit of {request_limit} requests per {interval} seconds exceeded')

    return check_rate_limit


async def check_password_breached(
    pw: SecretBytes,
    threshold: int,
    *,
    field_name: str = 'password',
    error_message: str = 'This password is known to hackers! Please use a different password',
) -> None:
    """
    Check a password against https://haveibeenpwned.com/API/v2#SearchingPwnedPasswordsByRange
    to see if it's known to hackers
    """
    pw_hash = hashlib.sha1(pw.get_secret_value()).hexdigest()
    r = await glove.http.get(f'https://api.pwnedpasswords.com/range/{pw_hash[:5]}')
    UnexpectedResponse.check(r)

    hash_suffix = pw_hash[5:].upper()
    for line in filter(None, r.text.split('\r\n')):
        line_suffix, count_str = line.split(':', 1)
        if line_suffix == hash_suffix:
            if int(count_str) > threshold:
                raise manual_response_error(field_name, error_message)
            break


def _hash_password(password: SecretBytes) -> str:
    return bcrypt.hashpw(password.get_secret_value(), bcrypt.gensalt(glove.settings.bcrypt_rounds)).decode()


async def hash_password(password: SecretBytes) -> str:
    """
    Hash a password to save in the db, run in a thread executor, I've checked that bcrypt releases the GIL
    """
    return await asyncio.get_running_loop().run_in_executor(None, _hash_password, password)


@lru_cache
def _get_dummy_hash() -> bytes:
    """
    This is only used to run checkpw when no password is available to prevent timing attack, however to guard
    against accidental usage, we hash a random byte string.
    """
    return bcrypt.hashpw(secrets.token_urlsafe().encode(), bcrypt.gensalt(glove.settings.bcrypt_rounds))


def _check_password(password: SecretBytes, expected_hash: Optional[str]) -> bool:
    """
    I've checked and bcrypt releases the GIL
    """
    if expected_hash is None:
        # this should never actually be used to check if a password is correct, just to avoid timing attack
        # by running checkpw on all login attempts
        bcrypt.checkpw(password.get_secret_value(), _get_dummy_hash())
        # we return False anyway to be sure
        return False
    else:

        return bcrypt.checkpw(password.get_secret_value(), expected_hash.encode())


async def check_password_correct(password: SecretBytes, expected_hash: Optional[str]) -> bool:
    """
    Check if a password is valid, run in a thread executor, I've checked that bcrypt releases the GIL
    """
    return await asyncio.get_running_loop().run_in_executor(None, _check_password, password, expected_hash)



import json
from typing import Any, Dict, Iterable, Optional, Type, TypeVar, Union

from httpx import Response as HttpxResponse
from starlette.requests import Request
from starlette.responses import JSONResponse, RedirectResponse

try:
    from devtools import pformat
except ImportError:  # pragma: no cover
    from pprint import pformat

__all__ = (
    'HttpRedirect',
    'redirect_handler',
    'HttpMessageError',
    'HttpOk',
    'HttpCreated',
    'HttpAccepted',
    'HttpBadRequest',
    'HttpUnauthorized',
    'HttpPaymentRequired',
    'HttpForbidden',
    'HttpNotFound',
    'HttpMethodNotAllowed',
    'HttpConflict',
    'HttpUnprocessableEntity',
    'HttpTooManyRequests',
    'Http470',
    'manual_response_error',
    'UnexpectedResponse',
)


class HttpRedirect(Exception):
    def __init__(self, location: str, *, status: int = 302):
        super().__init__(f'{status} redirect to {location!r}')
        self.status = status
        self.location = location


async def redirect_handler(request: Request, exc: HttpRedirect):
    return RedirectResponse(exc.location, status_code=exc.status)


class HttpMessageError(Exception):
    status: int
    custom_reason: str

    def __init__(self, message: str, *, details: Any = None, headers: Optional[Dict[str, str]] = None):
        super().__init__(message)
        self.message = message
        self.details = details
        self.headers = headers

    def __repr__(self) -> str:
        s = f'{self.__class__.__name__}({self.status}): {self.message}'
        if self.details:
            s += f'\ndetails:\n{pformat(self.details)}'
        return s

    def __str__(self) -> str:
        return repr(self)

    @staticmethod
    def handle(exc: 'HttpMessageError'):
        content = {}
        if exc.message:
            content['message'] = exc.message
        if exc.details:
            content['details'] = exc.details
        return JSONResponse(status_code=exc.status, content=content, headers=exc.headers)


class HttpOk(HttpMessageError):
    status = 200


class HttpCreated(HttpMessageError):
    status = 201


class HttpAccepted(HttpMessageError):
    status = 202


class HttpBadRequest(HttpMessageError):
    status = 400


class HttpUnauthorized(HttpMessageError):
    status = 401


class HttpPaymentRequired(HttpMessageError):
    status = 402


class HttpForbidden(HttpMessageError):
    status = 403


class HttpNotFound(HttpMessageError):
    status = 404


class HttpMethodNotAllowed(HttpMessageError):
    status = 405

    def __init__(self, message, allowed_methods, *, headers=None):
        headers = headers or {}
        headers.setdefault('Allow', ','.join(allowed_methods))
        super().__init__(message, details={'allowed_methods': allowed_methods}, headers=headers)


class HttpConflict(HttpMessageError):
    status = 409


class HttpUnprocessableEntity(HttpMessageError):
    status = 422


class HttpTooManyRequests(HttpMessageError):
    status = 429


class Http470(HttpMessageError):
    status = 470
    custom_reason = 'Invalid user input'


ExcType = TypeVar('ExcType', bound=HttpMessageError)
FieldType = Union[str, Iterable[str]]


def manual_response_error(
    field: 'FieldType', msg: str, exc: Type[ExcType] = HttpUnprocessableEntity, *, error_location: str = 'body'
) -> ExcType:
    """
    Build error details that reflect how fastapi/pydantic structures errors so the frontend
    (and reactstrap_toolbox/requests.js in particular) can show user errors easily.
    """
    loc = [error_location]
    if isinstance(field, str):
        loc.append(field)
    else:
        loc += list(field)
    return exc('validation error', details=[{'loc': loc, 'msg': msg}])


class UnexpectedResponse(ValueError):
    def __init__(self, r: HttpxResponse):
        self.status_code = r.status_code
        self.response = r
        super().__init__(f'{r.request.method} {r.request.url}, unexpected response: {r.status_code}')

    @classmethod
    def check(cls, r: HttpxResponse, *, allowed_responses: Iterable[int] = (200, 201)) -> None:
        if r.status_code not in allowed_responses:
            raise cls(r)

    def __repr__(self):
        try:
            self.body = self.response.json()
        except ValueError:
            self.body = b = self.response.text
        else:
            b = json.dumps(self.body, indent=2)
        return f'UnexpectedResponse("{self.args[0]}:\n{b}")'



from typing import TYPE_CHECKING, Any, Dict, Optional

from starlette.testclient import TestClient

try:
    from devtools.prettier import pformat
except ImportError:
    from pprint import pformat

try:
    import pytest
except ImportError:
    pytest = None

if TYPE_CHECKING:
    from requests import Response


class Client(TestClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_request: Optional['Response'] = None

    def request(self, *args, **kwargs) -> 'Response':
        r = super().request(*args, **kwargs)
        self.last_request = r
        return r

    def get_json(
        self,
        url: str,
        *,
        allow_redirects: bool = False,
        status: Optional[int] = 200,
        headers: Dict[str, str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        r = self.get(url, allow_redirects=allow_redirects, headers=headers, **kwargs)
        check_response(r, status)
        return r.json()

    def post_json(
        self,
        url: str,
        json: Any = None,
        *,
        allow_redirects: bool = False,
        status: Optional[int] = 200,
        headers: Dict[str, str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        r = self.post(url, json=json, allow_redirects=allow_redirects, headers=headers, **kwargs)
        check_response(r, status)
        return r.json()


def check_response(response: 'Response', expected_status: Optional[int]) -> None:
    if expected_status is None or response.status_code == expected_status:
        return

    try:
        body = pformat(response.json())
    except ValueError:
        body = response.text

    req = response.request
    msg = (
        f'{req.method} {req.url} returned unexpected status: {response.status_code} (expected {expected_status}), '
        f'body:\n{body}'
    )

    if pytest:  # pragma: no branch
        pytest.fail(msg)
    else:  # pragma: no cover
        raise ValueError(msg)



import asyncio
import os
from typing import Literal

import arq
import httpx
from buildpg.asyncpg import BuildPgPool
from pydantic.env_settings import BaseSettings as PydanticBaseSettings
from uvicorn.importer import ImportFromStringError, import_from_string

from .db import create_pg_pool
from .settings import BaseSettings

__all__ = ('glove',)


class Glove:
    _settings: BaseSettings
    _http: httpx.AsyncClient
    pg: BuildPgPool
    redis: arq.ArqRedis

    async def startup(self, *, run_migrations: Literal[True, False, 'unless-test-mode'] = 'unless-test-mode') -> None:
        from .logs import setup_sentry

        setup_sentry()

        if run_migrations == 'unless-test-mode':
            run_migrations = not self.settings.test_mode

        if not hasattr(self, 'pg'):
            self.pg = await create_pg_pool(self.settings, run_migrations=run_migrations)
        if not hasattr(self, 'redis') and self.settings.redis_settings:
            self.redis = await arq.create_pool(self.settings.redis_settings)

    def context(self) -> 'GloveContext':
        return GloveContext(self)

    async def shutdown(self) -> None:
        coros = []
        if pg := getattr(self, 'pg', None):
            coros.append(pg.close())
        if http := getattr(self, '_http', None):
            coros.append(http.aclose())
        if redis := getattr(self, 'redis', None):
            redis.close()
            coros.append(redis.wait_closed())
        await asyncio.gather(*coros)
        for prop in 'pg', '_http', 'redis':
            if hasattr(self, prop):
                delattr(self, prop)

    @property
    def http(self) -> httpx.AsyncClient:
        http = getattr(self, '_http', None)
        if http is None:
            http = self._http = httpx.AsyncClient(timeout=self.settings.http_client_timeout)
        return http

    @property
    def settings(self) -> BaseSettings:
        settings = getattr(self, '_settings', None)
        if settings is None:
            settings_path = os.environ['foxglove_settings_path']
            try:
                settings_cls = import_from_string(settings_path)
            except ImportFromStringError as exc:
                raise ImportError(f'unable to import "{settings_path}", {exc.__class__.__name__}: {exc}')

            if not isinstance(settings_cls, type) or not issubclass(settings_cls, PydanticBaseSettings):
                raise ImportError(f'settings "{settings_cls}" (from "{settings_path}"), is not a valid Settings class')

            settings = self._settings = settings_cls()
        return settings


class GloveContext:
    def __init__(self, g: Glove):
        self._glove = g

    async def __aenter__(self) -> None:
        await self._glove.startup()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self._glove.shutdown()


glove = Glove()



import asyncio
import logging

from starlette.endpoints import WebSocketEndpoint
from starlette.responses import Response
from starlette.routing import Route, WebSocketRoute
from watchgod import DefaultWatcher, awatch

logger = logging.getLogger('foxglove.cli')

extra_ignore = r'\.py$', r'\.pyx$', r'\.pyd$', r'\.s[ca]ss$', r'sasstastic\.yml$'


class FoxgloveWatcher(DefaultWatcher):
    ignored_file_regexes = DefaultWatcher.ignored_file_regexes + extra_ignore


def devtools_up(request):
    return Response(b'server up\n', media_type='text/plain')


def reload_endpoint(watch_path: str):
    async def watch_reload(prompt_reload):
        async for _ in awatch(watch_path, watcher_cls=FoxgloveWatcher):
            await prompt_reload()

    class ReloadWs(WebSocketEndpoint):
        encoding = 'text'

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._watch_task = asyncio.create_task(watch_reload(self.prompt_reload))
            self.ws = None

        async def prompt_reload(self):
            if self.ws:
                logger.debug('prompting reload')
                await self.ws.send_text('reload')

        async def on_connect(self, websocket):
            logger.debug('reload websocket connecting')
            await websocket.accept()
            self.ws = websocket

        async def on_disconnect(self, websocket, close_code):
            logger.debug('reload websocket disconnecting')
            self._watch_task.cancel()
            try:
                await self._watch_task
            except asyncio.CancelledError:
                logger.debug('file watcher cancelled')

    return [
        WebSocketRoute('/.devtools/reload-ws/', ReloadWs, name='devtools-reload'),
        Route('/.devtools/up/', devtools_up, methods=['GET']),
    ]



import asyncio
import logging
import os

from buildpg.asyncpg import BuildPgPool, DuplicateDatabaseError, UniqueViolationError, create_pool_b

from ..settings import BaseSettings
from .utils import AsyncPgContext, lenient_conn

logger = logging.getLogger('foxglove.db')
__all__ = 'create_pg_pool', 'prepare_database', 'reset_database'


async def create_pg_pool(settings: BaseSettings, *, run_migrations: bool = True) -> BuildPgPool:
    await prepare_database(settings, False, run_migrations=run_migrations)
    return await create_pool_b(
        settings.pg_dsn,
        min_size=settings.pg_pool_min_size,
        max_size=settings.pg_pool_max_size,
    )


async def prepare_database(settings: BaseSettings, overwrite_existing: bool, *, run_migrations: bool = True) -> bool:
    db_created = await create_database(settings, overwrite_existing)
    if settings.pg_migrations and run_migrations:
        from .migrations import run_migrations as run_migrations_
        from .patches import import_patches

        patches = import_patches(settings)

        await run_migrations_(settings, patches, True)
    return db_created


async def create_database(settings: BaseSettings, overwrite_existing: bool) -> bool:  # noqa: C901 (ignore complexity)
    """
    (Re)create a fresh database and run migrations.
    :param settings: settings to use for db connection
    :param overwrite_existing: whether or not to drop an existing database if it exists
    :return: whether or not a database has been (re)created
    """
    if settings.pg_db_exists:
        conn = await lenient_conn(settings, with_db=True)
        try:
            tables = await conn.fetchval("select count(*) from information_schema.tables where table_schema='public'")
            logger.info('existing tables: %d', tables)
            if tables > 0:
                if overwrite_existing:
                    logger.debug('database already exists...')
                else:
                    logger.debug('database already exists ✓')
                    return False
        finally:
            await conn.close()
    else:
        conn = await lenient_conn(settings, with_db=False)
        try:
            if not overwrite_existing:
                # don't drop connections and try creating a db if it already exists and we're not overwriting
                exists = await conn.fetchval('select 1 from pg_database where datname=$1', settings.pg_name)
                if exists:
                    logger.info('database already exists ✓')
                    return False

            await conn.execute(
                """
                select pg_terminate_backend(pg_stat_activity.pid)
                from pg_stat_activity
                where pg_stat_activity.datname = $1 AND pid <> pg_backend_pid();
                """,
                settings.pg_name,
            )
            logger.debug('attempting to create database "%s"...', settings.pg_name)
            try:
                await conn.execute(f'create database {settings.pg_name}')
            except (DuplicateDatabaseError, UniqueViolationError):
                if overwrite_existing:
                    logger.debug('database already exists...')
                else:
                    logger.debug('database already exists, skipping creation')
                    return False
            else:
                logger.debug('database did not exist, now created')

            logger.debug('settings db timezone to utc...')
            await conn.execute(f"alter database {settings.pg_name} set timezone to 'UTC';")
        finally:
            await conn.close()

    logger.debug('dropping and re-creating teh schema...')
    async with AsyncPgContext(settings.pg_dsn) as conn:
        async with conn.transaction():
            await conn.execute('drop schema public cascade;\ncreate schema public;')

    logger.debug('creating tables from model definition...')
    async with AsyncPgContext(settings.pg_dsn) as conn:
        async with conn.transaction():
            await conn.execute(settings.sql)
    logger.info('database successfully setup ✓')
    return True


def reset_database(settings: BaseSettings):
    if not (os.getenv('CONFIRM_DATABASE_RESET') == 'confirm' or input('Confirm database reset? [yN] ') == 'y'):
        logger.info('cancelling')
    else:
        logger.info('resetting database...')
        asyncio.run(prepare_database(settings, True, run_migrations=False))
        logger.info('done.')
